package scws;

public class ShipCollisionWarningSystem {
	private static final String NO_WARNING    = "";
	private static final String WARNING_WATCH = "Ship is on collision course, have priority to go";
	private static final String WARNING_SLOW  = "Ship is on collision course, slow down and give way";
	private static final double SAFEZONE = 3.5;
	private Ship ship;
	public ShipCollisionWarningSystem(Ship ship) {
		this.ship = ship;
	}
	private boolean isHeadingExactlyTheSameDirection(int angle) {
	    return angle % 360 == 0;
	}
	private boolean isHeadingExactlyTheOppositeDirection(int angle) {
	    return angle % 360 == 180;
	}
	private boolean isHeadingAboutTheSameDirection(int angle) {
	    return 270 < (angle % 360) && (angle % 360) < 90;
	}
	private boolean isHeadingAboutTheOppositeDirection(int angle) {
	    return 90 < (angle % 360) && (angle % 360) < 270;
	}
	private boolean isHeadingToTheLeft(int angle) {
	    return 180 < (angle % 360);
	}
	private boolean isHeadingToTheRight(int angle) {
	    return 0 < (angle % 360) && (angle % 360) < 180;
	}
	private boolean isOnTheLeft(Position pos) {
		return Length.ZERO.compareTo(pos.x()) > 0;
	}
	private boolean isOnTheRight(Position pos) {
		return Length.ZERO.compareTo(pos.x()) < 0;
	}
	private boolean isBehindMe(Position pos) {
		return pos.y().compareTo(Length.ZERO) > 0;
	}
	private boolean isInFrontOfMe(Position pos) {
		return pos.y().compareTo(Length.ZERO) < 0;
	}
	private boolean isFasterThanMe(Ship ship) {
		return this.ship.speed().compareTo(ship.speed()) < 0;
	}
	private boolean isSlowerThanMe(Ship ship) {
		return this.ship.speed().compareTo(ship.speed()) > 0;
	}
	private boolean isHeavierThanMe(Ship ship) {
		return this.ship.mass().compareTo(ship.mass()) < 0;
	}
	private boolean isLighterThanMe(Ship ship) {
		return this.ship.mass().compareTo(ship.mass()) > 0;
	}
	/**
	 * Checks whether another ship is on collision course with us.
	 *
	 * @param pos    relative position of the other ship: our ship is at
	 *               (0,0) and heading towards (0,1) by definition;
	 *               (-1,0) is on our left, (1,0) is on our right,
	 *               (0,-1) is befind us.
	 * @param angle  relative direction of the other ship: we are
	 *               heading to 0° (north, 12'o clock) by definition;
	 *               90° is to the right (east, 3'o clock), 180° is the
	 *               opposite direction (south, 6'o clock), 270° is to
	 *               the left (wast, 9'o clock).
	 * @param ship   the other ship (length, width, speed, weight)
	 */
	public String checkCollision(Position pos, int angle, Ship ship) throws InvalidUnitException {
		if ((isOnTheLeft(pos) && isHeadingToTheLeft(angle)) ||
			(isOnTheRight(pos) && isHeadingToTheRight(angle))) {
			return NO_WARNING;
		}
		if (isHeadingExactlyTheSameDirection(angle) || isHeadingExactlyTheOppositeDirection(angle)) {
			Length min = null;
			if (this.ship.width().compareTo(ship.width()) < 0) {
				min = this.ship.width().div(2.0).add(ship.width().mul(SAFEZONE));
			} else {
				min = ship.width().div(2.0).add(this.ship.width().mul(SAFEZONE));
			}
			if (min.compareTo(pos.x()) < 1) {
				return NO_WARNING;
			}
			if (isHeadingExactlyTheSameDirection(angle)) {
				if (isBehindMe(pos)) {
					if (isFasterThanMe(ship)) {
						return WARNING_WATCH;
					}
				} else {
					if (isSlowerThanMe(ship)) {
						return WARNING_SLOW;
					}
				}
			} else {
				if (isHeavierThanMe(ship)) {
					return WARNING_SLOW;
				} else {
					return WARNING_WATCH;
				}
			}
			return NO_WARNING;
		}
		double ang = (90 - angle) * Math.PI / 180.0;;  // deg to rad, N is 0°, clockwise is positive
		Speed vx = ship.speed().mul(Math.cos(ang)); 
		Speed vy = ship.speed().mul(Math.sin(ang));
		Time shipTimeToCross            = pos.x().div(vx).neg();
		Time shipTimeRangeToPassThrough = ship.length().div(ship.speed()).mul(SAFEZONE).abs();
		Time shipTimeRangeForSelfWidth  = this.ship.width().div(Math.sin(ang)).div(ship.speed()).mul(SAFEZONE).abs();
		Time shipTimeWindowRange        = shipTimeRangeForSelfWidth.add(shipTimeRangeToPassThrough);
		Interval<Time> shipWindow = new Interval<Time>(shipTimeToCross.sub(shipTimeWindowRange), shipTimeToCross.add(shipTimeWindowRange));
		Length selfCrossPointDistance = pos.y().add(vy.mul(shipTimeToCross));
		Time selfTimeToCross            = selfCrossPointDistance.div(this.ship.speed());
		Time selfTimeRangeToPassThrough = this.ship.length().div(this.ship.speed()).mul(SAFEZONE).abs();
		Time selfTimeRangeForShipWidth  = ship.width().div(Math.sin(ang)).div(this.ship.speed()).mul(SAFEZONE).abs();
		Time selfTimeWindowRange        = selfTimeRangeForShipWidth.add(selfTimeRangeToPassThrough);
		Interval<Time> selfWindow = new Interval<Time>(selfTimeToCross.sub(selfTimeWindowRange), selfTimeToCross.add(selfTimeWindowRange));
		if (!shipWindow.hasIntersection(selfWindow)) {
			return NO_WARNING;
		}
		if (!isHeadingAboutTheOppositeDirection(angle) && isOnTheLeft(pos)) {
			return WARNING_WATCH;
		}
		if (!isHeadingAboutTheOppositeDirection(angle) && isOnTheRight(pos)) {
			return WARNING_SLOW;
		}
		if (isLighterThanMe(ship)) {
			return WARNING_WATCH;
		}
		return WARNING_SLOW;
	}
}
